/*
 * Copyright (C) 2023 Hasl Developers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <https://www.gnu.org/licenses/>.
 */

#include <glib/gi18n-lib.h>

#include <stringprep.h>

#include "haslgs2header.h"

#include "haslenums.h"

enum {
	PROP_0,
	PROP_STANDARD_MECHANISM,
	PROP_CHANNEL_BINDING,
	PROP_CHANNEL_BINDING_NAME,
	PROP_AUTHZID,
	N_PROPERTIES,
};
static GParamSpec *properties[N_PROPERTIES] = {NULL, };

struct _HaslGs2Header {
	GObject parent;

	gboolean standard_mechanism;

	HaslGs2HeaderChannelBinding channel_binding;
	char *channel_binding_name;

	char *authzid;
};

G_DEFINE_TYPE(HaslGs2Header, hasl_gs2_header, G_TYPE_OBJECT)

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
static void
hasl_gs2_header_get_property(GObject *obj, guint param_id, GValue *value,
                             GParamSpec *pspec)
{
	HaslGs2Header *header = HASL_GS2_HEADER(obj);

	switch(param_id) {
	case PROP_STANDARD_MECHANISM:
		g_value_set_boolean(value,
		                    hasl_gs2_header_get_standard_mechanism(header));
		break;
	case PROP_CHANNEL_BINDING:
		g_value_set_enum(value, hasl_gs2_header_get_channel_binding(header));
		break;
	case PROP_CHANNEL_BINDING_NAME:
		g_value_set_string(value,
		                   hasl_gs2_header_get_channel_binding_name(header));
		break;
	case PROP_AUTHZID:
		g_value_set_string(value, hasl_gs2_header_get_authzid(header));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
hasl_gs2_header_set_property(GObject *obj, guint param_id, const GValue *value,
                             GParamSpec *pspec)
{
	HaslGs2Header *header = HASL_GS2_HEADER(obj);

	switch(param_id) {
	case PROP_STANDARD_MECHANISM:
		hasl_gs2_header_set_standard_mechanism(header,
		                                       g_value_get_boolean(value));
		break;
	case PROP_CHANNEL_BINDING:
		hasl_gs2_header_set_channel_binding(header, g_value_get_enum(value));
		break;
	case PROP_CHANNEL_BINDING_NAME:
		hasl_gs2_header_set_channel_binding_name(header,
		                                         g_value_get_string(value));
		break;
	case PROP_AUTHZID:
		hasl_gs2_header_set_authzid(header, g_value_get_string(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
hasl_gs2_header_init(HaslGs2Header *header) {
	header->standard_mechanism = TRUE;
	header->channel_binding = HaslGs2HeaderChannelBindingClientNotSupported;
}

static void
hasl_gs2_header_finalize(GObject *obj) {
	HaslGs2Header *header = HASL_GS2_HEADER(obj);

	g_clear_pointer(&header->channel_binding_name, g_free);
	g_clear_pointer(&header->authzid, g_free);

	G_OBJECT_CLASS(hasl_gs2_header_parent_class)->finalize(obj);
}

static void
hasl_gs2_header_class_init(HaslGs2HeaderClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);

	obj_class->finalize = hasl_gs2_header_finalize;
	obj_class->get_property = hasl_gs2_header_get_property;
	obj_class->set_property = hasl_gs2_header_set_property;

	/**
	 * HaslGs2Header:standard-mechanism:
	 *
	 * Whether or not the mechanism is a standard GSS-API mechanism.
	 *
	 * Since: 0.3
	 */
	properties[PROP_STANDARD_MECHANISM] = g_param_spec_boolean(
		"standard-mechanism", NULL, NULL,
		TRUE,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * HaslGs2Header:channel-binding:
	 *
	 * The channel binding flag for this header.
	 *
	 * Since: 0.3
	 */
	properties[PROP_CHANNEL_BINDING] = g_param_spec_enum(
		"channel-binding", NULL, NULL,
		HASL_TYPE_GS2_HEADER_CHANNEL_BINDING,
		HaslGs2HeaderChannelBindingClientNotSupported,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * HaslGs2Header:channel-binding-name:
	 *
	 * The name of the channel binding data to use.
	 *
	 * If [property@Gs2Header:channel-binding] is set to ClientSupported, this
	 * must not be %NULL.
	 *
	 * Since: 0.3
	 */
	properties[PROP_CHANNEL_BINDING_NAME] = g_param_spec_string(
		"channel-binding-name", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);


	/**
	 * HaslGs2Header:authzid:
	 *
	 * An optional authzid to use. This is the username to authenticate as when
	 * using admin/moderator credentials.
	 *
	 * Since: 0.3
	 */
	properties[PROP_AUTHZID] = g_param_spec_string(
		"authzid", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	g_object_class_install_properties(obj_class, N_PROPERTIES, properties);
}

/******************************************************************************
 * Public API
 *****************************************************************************/
HaslGs2Header *
hasl_gs2_header_new(void) {
	return g_object_new(HASL_TYPE_GS2_HEADER, NULL);
}

char *
hasl_gs2_header_to_string(HaslGs2Header *header, GError **error) {
	GString *str = NULL;

	g_return_val_if_fail(HASL_IS_GS2_HEADER(header), NULL);

	str = g_string_new("");

	if(!header->standard_mechanism) {
		g_string_append(str, "F,");
	}

	/* Add our channel binding information. */
	if(header->channel_binding == HaslGs2HeaderChannelBindingClientSupported) {
		gboolean invalid = FALSE;

		if(header->channel_binding_name == NULL ||
		   header->channel_binding_name[0] == '\0')
		{
			g_set_error_literal(error, HASL_GS2_HEADER_DOMAIN, 0,
			                    "channel binding was set to supported but no "
			                    "name was given.");

			g_string_free(str, TRUE);

			return NULL;
		}

		/* Look to see if we have any invalid characters in the name. */
		invalid = g_regex_match_simple("[^A-Za-z0-9.-]",
		                               header->channel_binding_name,
		                               G_REGEX_DEFAULT, G_REGEX_MATCH_DEFAULT);
		if(invalid) {
			g_set_error(error, HASL_GS2_HEADER_DOMAIN, 0,
			            "channel binding name \"%s\" contains invalid "
			            "characters. Only alphanumeric . and - are supported",
			            header->channel_binding_name);

			g_string_free(str, TRUE);

			return NULL;
		}

		g_string_append_printf(str, "p=%s", header->channel_binding_name);
	} else if(header->channel_binding == HaslGs2HeaderChannelBindingClientNotSupported)
	{
		g_string_append(str, "n");
	} else if(header->channel_binding == HaslGs2HeaderChannelBindingClientSupportedServerUnknown)
	{
		g_string_append(str, "y");
	} else {
		g_set_error(error, HASL_GS2_HEADER_DOMAIN, 0,
		            "unknown channel binding value %d",
		            header->channel_binding);

		g_string_free(str, TRUE);

		return NULL;
	}
	g_string_append(str, ",");

	/* Add the authzid if we have it. */
	if(header->authzid != NULL && header->authzid[0] != '\0') {
		char buffer[1024];
		int res = 0;

		g_strlcpy(buffer, header->authzid, sizeof(buffer));
		res = stringprep(buffer, sizeof(buffer), 0, stringprep_saslprep);
		if(res != STRINGPREP_OK) {
			g_set_error(error, HASL_GS2_HEADER_DOMAIN, 0,
			            _("stringprep failed: %s"), stringprep_strerror(res));

			g_string_free(str, TRUE);

			return NULL;
		}

		g_string_append_printf(str, "a=%s", header->authzid);
	}
	g_string_append(str, ",");

	return g_string_free_and_steal(str);
}

gboolean
hasl_gs2_header_get_standard_mechanism(HaslGs2Header *header) {
	g_return_val_if_fail(HASL_IS_GS2_HEADER(header), TRUE);

	return header->standard_mechanism;
}

void
hasl_gs2_header_set_standard_mechanism(HaslGs2Header *header,
                                       gboolean standard)
{
	g_return_if_fail(HASL_IS_GS2_HEADER(header));

	if(header->standard_mechanism != standard) {
		header->standard_mechanism = standard;

		g_object_notify_by_pspec(G_OBJECT(header),
		                         properties[PROP_STANDARD_MECHANISM]);
	}
}

HaslGs2HeaderChannelBinding
hasl_gs2_header_get_channel_binding(HaslGs2Header *header) {
	g_return_val_if_fail(HASL_IS_GS2_HEADER(header),
	                     HaslGs2HeaderChannelBindingClientNotSupported);

	return header->channel_binding;
}

void
hasl_gs2_header_set_channel_binding(HaslGs2Header *header,
                                    HaslGs2HeaderChannelBinding channel_binding)
{
	g_return_if_fail(HASL_IS_GS2_HEADER(header));

	if(header->channel_binding != channel_binding) {
		header->channel_binding = channel_binding;

		g_object_notify_by_pspec(G_OBJECT(header),
		                         properties[PROP_CHANNEL_BINDING]);
	}
}

const char *
hasl_gs2_header_get_channel_binding_name(HaslGs2Header *header) {
	g_return_val_if_fail(HASL_IS_GS2_HEADER(header), NULL);

	return header->channel_binding_name;
}

void
hasl_gs2_header_set_channel_binding_name(HaslGs2Header *header,
                                         const char *name)
{
	g_return_if_fail(HASL_IS_GS2_HEADER(header));

	if(g_strcmp0(header->channel_binding_name, name) != 0) {
		g_free(header->channel_binding_name);
		header->channel_binding_name = g_strdup(name);

		g_object_notify_by_pspec(G_OBJECT(header),
		                         properties[PROP_CHANNEL_BINDING_NAME]);
	}
}

const char *
hasl_gs2_header_get_authzid(HaslGs2Header *header) {
	g_return_val_if_fail(HASL_IS_GS2_HEADER(header), NULL);

	return header->authzid;
}

void
hasl_gs2_header_set_authzid(HaslGs2Header *header, const char *authzid) {
	g_return_if_fail(HASL_IS_GS2_HEADER(header));

	if(g_strcmp0(header->authzid, authzid) != 0) {
		g_free(header->authzid);
		header->authzid = g_strdup(authzid);

		g_object_notify_by_pspec(G_OBJECT(header), properties[PROP_AUTHZID]);
	}
}
